from .scope_maps import ScopeMaps
from ..utils import NewCentralURLs
import copy

scope_maps = ScopeMaps()
urls = NewCentralURLs()

SUPPORTED_SCOPES = ["SITE", "SITE_COLLECTION", "DEVICE", "DEVICE_GROUP"]
DEFAULT_LIMIT = 100


def fetch_attribute(obj, attribute):
    """
    This function fetches the value associated with the provided attribute in the object, if it exists.

    :param obj: Object whose attribute has to be returned
    :type obj: class
    :param attribute: Attribute within the object that has to be returned
    :type attribute: str

    :return: Value of the required attribute, if it exists in the object. If the attribute doesn't exist, it will return None.
    :rtype: string
    """
    if hasattr(obj, attribute):
        return getattr(obj, attribute)
    return None


def update_attribute(obj, attribute, new_value):
    """
    This function updates the value of the provided attribute in the object, if it exists.

    :param obj: Object whose attribute has to be updated
    :type obj: class
    :param attribute: Attribute within the object that has to be updated
    :type attribute: str

    :return: True if the attribute was successfully updated, else False.
    :rtype: bool
    """
    if hasattr(obj, attribute):
        setattr(obj, attribute, new_value)
        return True
    return False


def get_attributes(obj):
    """
    This function returns all attributes of the provided class

    :param obj: Object whose attribute has to be returned
    :type obj: class

    :return: Returns attributes defined in the Site Collection object
    :rtype: dict
    """
    return {k: v for k, v in obj.__dict__.items() if not callable(v)}


def assign_profile(obj, profile_name, profile_persona):
    """
    Performs a POST call to associate a profile(with the provided name & persona) to the scope

    :param obj: Class instance to which profile has to be assigned
    :type obj: class
    :param profile_name: Name of profile that has to be assigned
    :type profile_name: str
    :param profile_persona: Device persona of the profile that has to be assigned
    :type profile_persona: str

    :return: True if the profile assignment to scope was successful, else False
    :rtype: bool
    """
    resp = scope_maps.associate_profile_to_scope(
        central_conn=obj.central_conn,
        scope_id=obj.get_id(),
        profile_name=profile_name,
        persona=profile_persona,
    )
    if resp["code"] == 200:
        add_profile(obj=obj, name=profile_name, persona=profile_persona)
        return True
    else:
        obj.central_conn.logger.error(
            "Unable to assign profile "
            + profile_name
            + " to site collection "
            + obj.get_name()
        )
        return False


def unassign_profile(obj, profile_name, profile_persona):
    """
    Performs a DELETE call to unassign a profile(with the provided name & persona) from the scope.

    :param obj: Class instance to which profile has to be unassigned
    :type obj: class
    :param profile_name: Name of profile that has to be unassigned
    :type profile_name: str
    :param profile_persona: Device persona of the profile that has to be unassigned
    :type profile_persona: str

    :return: True if the profile removal to scope was successful, else False
    :rtype: bool
    """
    resp = scope_maps.unassociate_profile_from_scope(
        central_conn=obj.central_conn,
        scope_id=obj.get_id(),
        profile_name=profile_name,
        persona=profile_persona,
    )
    if resp["code"] == 200:
        remove_profile(obj=obj, name=profile_name, persona=profile_persona)
        return True
    else:
        obj.central_conn.logger.error(
            "Unable to unassign profile "
            + profile_name
            + " to site collection "
            + obj.get_name()
        )
        return False


def add_profile(obj, name, persona):
    """
    Adds the dictionary matching the profile name and device persona to the assigned profile attribute in self

    :param obj: Class instance to which profile has to be added
    :type obj: class
    :param name: Name of profile that has to be added
    :type name: str
    :param persona: Device persona of profile that has to be added
    :type persona: str
    """
    obj.assigned_profiles.append({"persona": persona, "resource": name})


def remove_profile(obj, name, persona):
    """
    Removes the dictionary matching the profile name and device persona from the assigned profile attribute in self

    :param obj: Class instance to which profile has to be removed
    :type obj: class
    :param name: Name of profile that has to be removed
    :type name: str
    :param persona: Device persona of profile that has to be removed
    :type persona: str

    :return: True if the matching name and persona was found and removed from assigned profiles, else returns False
    :rtype: bool
    """
    remove_status = False
    index = None
    for id_element, element in enumerate(obj.assigned_profiles):
        if element["persona"] == persona and element["resource"] == name:
            index = id_element
            break
    if index is not None:
        obj.assigned_profiles.pop(index)
        remove_status = True
    return remove_status


def get_all_scope_elements(obj, scope):
    """
    This function makes GET API calls to Central to get all the elements of the specified scope. This method is supported for site, site collection, device & device groups scopes.

    :param obj: Class instance that will be used to make API calls to Central
    :type obj: class
    :param scope: The type of the element. SITE_COLLECTION, SITE, DEVICE, DEVICE_GROUP are valid parameters for this argument
    :type scope: str

    :return: List of all scope elements. If there are errors it will return None.
    :rtype: list
    """
    if scope not in SUPPORTED_SCOPES:
        obj.central_conn.logger.error(
            "Unknown scope provided. Please provide one of the supported scopes - "
            ", ".join(SUPPORTED_SCOPES)
        )
        return None
    limit = DEFAULT_LIMIT
    offset = 1
    scope_elements = []
    number_of_scope_elements = None
    while (
        number_of_scope_elements is None
        or len(scope_elements) < number_of_scope_elements
    ):
        resp = get_scope_elements(obj, scope, limit=limit, offset=offset)
        if resp["code"] == 200:
            offset += 1
            resp_message = resp["msg"]
            if number_of_scope_elements is None:
                number_of_scope_elements = resp_message["total"]

            scope_elements.extend(
                [scope_element for scope_element in resp_message["items"]]
            )
        else:
            obj.central_conn.logger.error(resp["msg"]["message"])
        return scope_elements


def get_scope_elements(obj, scope, limit=50, offset=0, filter_field="", sort=""):
    """
    This function makes GET APIs to Central to get the list of scope elements based on the provided attributes. This method is supported for site, site collection, device & device groups scopes.

    :param obj: Class instance that will be used to make API calls to Central
    :type obj: class
    :param scope: The type of the element. SITE_COLLECTION, SITE, DEVICE, DEVICE_GROUP are valid parameters for this argument
    :type scope: str
    :param limit: Number of scope elements to be fetched, defaults to 50
    :type limit: int
    :param offset: Pagination start index, defaults to 1
    :type offset: int
    :param filter_field: Field that needs to be used for sorting the list of sites or site collections. Accepted values for sites is scopeName, address, state, country, city, deviceCount, collectionName, zipcode, timezone. Accepted values for this argument for site_collection is scopeName, description, deviceCount, siteCount
    :type filter_field: str, optional
    :param sort: Direction of sorting for the field. ASC or DESC are accepted values for this argument
    :type sort: str, optional

    :return: List of scope elements based on the provided arguments. If there are errors it will return None.
    :rtype: list
    """
    if scope not in SUPPORTED_SCOPES:
        obj.central_conn.logger.error(
            "Unknown scope provided. Please provide one of the supported scopes - "
            ", ".join(SUPPORTED_SCOPES)
        )
        return None

    api_path = urls.fetch_url("SCOPES", scope)
    api_method = "GET"
    api_params = {"limit": limit, "offset": offset}

    if filter_field:
        api_params["filter"] = filter_field
    if sort:
        api_params["sort"] = sort

    resp = obj.central_conn.command(
        api_method=api_method, api_path=api_path, api_params=api_params
    )
    return resp


def set_attributes(
    obj,
    attributes_dict,
    required_attributes,
    optional_attributes,
    object_attributes=None,
):
    """
    This sets the attributes of the given object based on the attributes_dict

    :param obj: Class instance that will be used to make API calls to Central
    :type obj: class
    :param attributes_dict: Attributes that will be updated to self
    :type attributes_dict: dict
    """
    for attr in required_attributes:
        value = attributes_dict[attr]
        setattr(obj, attr, value)

    for attr, default_value in optional_attributes.items():
        value = attributes_dict.get(attr)
        if not value:
            if isinstance(default_value, list):
                value = copy.deepcopy(default_value)
            else:
                value = default_value
        setattr(obj, attr, value)
    if object_attributes:
        for attr, default_value in object_attributes.items():
            if attr in attributes_dict:
                setattr(obj, attr, attributes_dict[attr])
            else:
                setattr(obj, attr, default_value)


def get_scope_element(obj, scope, scope_id):
    """
    This function makes GET APIs to Central to find the specified scope element & return that data. This method is supported for site, site collection, device & device groups scopes.

    :param obj: Class instance that will be used to make API calls to Central
    :type obj: class
    :param scope: The type of the element. The type of the element. SITE_COLLECTION, SITE, DEVICE, DEVICE_GROUP are valid parameters for this argument
    :type scope: str
    :param scope_id: ID of the scope element to be returned
    :type scope_id: int

    :return: If the site collection is found, the attributes associated with site collection is returned from Central, else None is returned
    :rtype: dict
    """
    scope_elements_list = get_all_scope_elements(obj=obj, scope=scope)
    for element in scope_elements_list:
        if element["scopeId"] == str(scope_id):
            return element
    return None

def rename_keys(api_dict, api_attribute_mapping):
        """
        This function renames the keys of the site attributes from the API
        response

        :param api_dict: dict of information from Central API Response
        :type api_dict: dict
        :param api_attribute_mapping: Dict of Object attributes mapped to keys
        from Central API Response
        :type api_attribute_mapping: dict

        :return: Renamed dictionary of Object attributes. The renamed keys maps
        to attributes that will be defined in the Object
        :rtype: dict
        """

        #️ Removing reduntant keys
        extra_keys = ["type", "scopeId"]
        for key in extra_keys:
            if key in api_dict.keys():
                del api_dict[key]
        integer_attributes = ["id", "collectionId", "deviceCount"]
        renamed_dict = {}
        for key, value in api_dict.items():
            new_key = api_attribute_mapping.get(key)
            if new_key:
                if key in integer_attributes and value:
                    value = int(value)
                elif key == "timezone" and "timezoneId" in value:
                    value = value["timezoneId"]
                renamed_dict[new_key] = value
            else:
                raise ValueError(f"Unknown attribute {key} found" +
                                 "in API response")
        return renamed_dict